@@ -21,6 +21,29 @@ protocol.types = {
   15: 'auth'
 }
 
+protocol.requiredHeaderFlags = {
+  1: 0, // 'connect'
+  2: 0, // 'connack'
+  4: 0, // 'puback'
+  5: 0, // 'pubrec'
+  6: 2, // 'pubrel'
+  7: 0, // 'pubcomp'
+  8: 2, // 'subscribe'
+  9: 0, // 'suback'
+  10: 2, // 'unsubscribe'
+  11: 0, // 'unsuback'
+  12: 0, // 'pingreq'
+  13: 0, // 'pingresp'
+  14: 0, // 'disconnect'
+  15: 0 // 'auth'
+}
+
+protocol.requiredHeaderFlagsErrors = {}
+for (const k in protocol.requiredHeaderFlags) {
+  const v = protocol.requiredHeaderFlags[k]
+  protocol.requiredHeaderFlagsErrors[k] = 'Invalid header flag bits, must be 0x' + v.toString(16) + ' for ' + protocol.types[k] + ' packet'
+}
+
 /* Mnemonic => Command code */
 protocol.codes = {}
 for (const k in protocol.types) {
@@ -39,6 +62,7 @@ protocol.RETAIN_MASK = 0x01
 /* Length */
 protocol.VARBYTEINT_MASK = 0x7F
 protocol.VARBYTEINT_FIN_MASK = 0x80
+protocol.VARBYTEINT_MAX = 268435455
 
 /* Connack */
 protocol.SESSIONPRESENT_MASK = 0x01
@@ -170,6 +194,8 @@ protocol.SUBACK_HEADER = Buffer.from([protocol.codes.suback << protocol.CMD_SHIF
 protocol.VERSION3 = Buffer.from([3])
 protocol.VERSION4 = Buffer.from([4])
 protocol.VERSION5 = Buffer.from([5])
+protocol.VERSION131 = Buffer.from([131])
+protocol.VERSION132 = Buffer.from([132])
 
 /* QoS */
 protocol.QOS = [0, 1, 2].map(qos => {
@@ -182,3 +208,83 @@ protocol.EMPTY = {
   pingresp: Buffer.from([protocol.codes.pingresp << 4, 0]),
   disconnect: Buffer.from([protocol.codes.disconnect << 4, 0])
 }
+
+protocol.MQTT5_PUBACK_PUBREC_CODES = {
+  0x00: 'Success',
+  0x10: 'No matching subscribers',
+  0x80: 'Unspecified error',
+  0x83: 'Implementation specific error',
+  0x87: 'Not authorized',
+  0x90: 'Topic Name invalid',
+  0x91: 'Packet identifier in use',
+  0x97: 'Quota exceeded',
+  0x99: 'Payload format invalid'
+}
+
+protocol.MQTT5_PUBREL_PUBCOMP_CODES = {
+  0x00: 'Success',
+  0x92: 'Packet Identifier not found'
+}
+
+protocol.MQTT5_SUBACK_CODES = {
+  0x00: 'Granted QoS 0',
+  0x01: 'Granted QoS 1',
+  0x02: 'Granted QoS 2',
+  0x80: 'Unspecified error',
+  0x83: 'Implementation specific error',
+  0x87: 'Not authorized',
+  0x8F: 'Topic Filter invalid',
+  0x91: 'Packet Identifier in use',
+  0x97: 'Quota exceeded',
+  0x9E: 'Shared Subscriptions not supported',
+  0xA1: 'Subscription Identifiers not supported',
+  0xA2: 'Wildcard Subscriptions not supported'
+}
+
+protocol.MQTT5_UNSUBACK_CODES = {
+  0x00: 'Success',
+  0x11: 'No subscription existed',
+  0x80: 'Unspecified error',
+  0x83: 'Implementation specific error',
+  0x87: 'Not authorized',
+  0x8F: 'Topic Filter invalid',
+  0x91: 'Packet Identifier in use'
+}
+
+protocol.MQTT5_DISCONNECT_CODES = {
+  0x00: 'Normal disconnection',
+  0x04: 'Disconnect with Will Message',
+  0x80: 'Unspecified error',
+  0x81: 'Malformed Packet',
+  0x82: 'Protocol Error',
+  0x83: 'Implementation specific error',
+  0x87: 'Not authorized',
+  0x89: 'Server busy',
+  0x8B: 'Server shutting down',
+  0x8D: 'Keep Alive timeout',
+  0x8E: 'Session taken over',
+  0x8F: 'Topic Filter invalid',
+  0x90: 'Topic Name invalid',
+  0x93: 'Receive Maximum exceeded',
+  0x94: 'Topic Alias invalid',
+  0x95: 'Packet too large',
+  0x96: 'Message rate too high',
+  0x97: 'Quota exceeded',
+  0x98: 'Administrative action',
+  0x99: 'Payload format invalid',
+  0x9A: 'Retain not supported',
+  0x9B: 'QoS not supported',
+  0x9C: 'Use another server',
+  0x9D: 'Server moved',
+  0x9E: 'Shared Subscriptions not supported',
+  0x9F: 'Connection rate exceeded',
+  0xA0: 'Maximum connect time',
+  0xA1: 'Subscription Identifiers not supported',
+  0xA2: 'Wildcard Subscriptions not supported'
+}
+
+protocol.MQTT5_AUTH_CODES = {
+  0x00: 'Success',
+  0x18: 'Continue authentication',
+  0x19: 'Re-authenticate'
+}
