@@ -54,9 +54,19 @@ class Parser extends EventEmitter {
   _parseHeader () {
     // There is at least one byte in the buffer
     const zero = this._list.readUInt8(0)
-    this.packet.cmd = constants.types[zero >> constants.CMD_SHIFT]
+    const cmdIndex = zero >> constants.CMD_SHIFT
+    this.packet.cmd = constants.types[cmdIndex]
+    const headerFlags = zero & 0xf
+    const requiredHeaderFlags = constants.requiredHeaderFlags[cmdIndex]
+    if (requiredHeaderFlags != null && headerFlags !== requiredHeaderFlags) {
+      // Where a flag bit is marked as “Reserved” in Table 2.2 - Flag Bits, it is reserved for future use and MUST be set to the value listed in that table [MQTT-2.2.2-1]. If invalid flags are received, the receiver MUST close the Network Connection [MQTT-2.2.2-2]
+      return this._emitError(new Error(constants.requiredHeaderFlagsErrors[cmdIndex]))
+    }
     this.packet.retain = (zero & constants.RETAIN_MASK) !== 0
     this.packet.qos = (zero >> constants.QOS_SHIFT) & constants.QOS_MASK
+    if (this.packet.qos > 2) {
+      return this._emitError(new Error('Packet must not have both QoS bits set to 1'))
+    }
     this.packet.dup = (zero & constants.DUP_MASK) !== 0
     debug('_parseHeader: packet: %o', this.packet)
 
@@ -173,16 +183,30 @@ class Parser extends EventEmitter {
       return this._emitError(new Error('Packet too short'))
     }
 
+    if (this._list.readUInt8(this._pos) & 0x1) {
+      // The Server MUST validate that the reserved flag in the CONNECT Control Packet is set to zero and disconnect the Client if it is not zero [MQTT-3.1.2-3]
+      return this._emitError(new Error('Connect flag bit 0 must be 0, but got 1'))
+    }
     // Parse connect flags
     flags.username = (this._list.readUInt8(this._pos) & constants.USERNAME_MASK)
     flags.password = (this._list.readUInt8(this._pos) & constants.PASSWORD_MASK)
     flags.will = (this._list.readUInt8(this._pos) & constants.WILL_FLAG_MASK)
 
+    const willRetain = !!(this._list.readUInt8(this._pos) & constants.WILL_RETAIN_MASK)
+    const willQos = (this._list.readUInt8(this._pos) &
+        constants.WILL_QOS_MASK) >> constants.WILL_QOS_SHIFT
+
     if (flags.will) {
       packet.will = {}
-      packet.will.retain = (this._list.readUInt8(this._pos) & constants.WILL_RETAIN_MASK) !== 0
-      packet.will.qos = (this._list.readUInt8(this._pos) &
-        constants.WILL_QOS_MASK) >> constants.WILL_QOS_SHIFT
+      packet.will.retain = willRetain
+      packet.will.qos = willQos
+    } else {
+      if (willRetain) {
+        return this._emitError(new Error('Will Retain Flag must be set to zero when Will Flag is set to 0'))
+      }
+      if (willQos) {
+        return this._emitError(new Error('Will QoS must be set to zero when Will Flag is set to 0'))
+      }
     }
 
     packet.clean = (this._list.readUInt8(this._pos) & constants.CLEAN_SESSION_MASK) !== 0
@@ -250,7 +274,11 @@ class Parser extends EventEmitter {
     const packet = this.packet
 
     if (this._list.length < 1) return null
-    packet.sessionPresent = !!(this._list.readUInt8(this._pos++) & constants.SESSIONPRESENT_MASK)
+    const flags = this._list.readUInt8(this._pos++)
+    if (flags > 1) {
+      return this._emitError(new Error('Invalid connack flags, bits 7-1 must be set to 0'))
+    }
+    packet.sessionPresent = !!(flags & constants.SESSIONPRESENT_MASK)
 
     if (this.settings.protocolVersion === 5) {
       if (this._list.length >= 2) {
@@ -307,10 +335,6 @@ class Parser extends EventEmitter {
     let nl
     let subscription
 
-    if (packet.qos !== 1) {
-      return this._emitError(new Error('Wrong subscribe header'))
-    }
-
     packet.subscriptions = []
 
     if (!this._parseMessageId()) { return }
@@ -323,6 +347,8 @@ class Parser extends EventEmitter {
       }
     }
 
+    if (packet.length <= 0) { return this._emitError(new Error('Malformed subscribe, no payload specified')) }
+
     while (this._pos < packet.length) {
       // Parse topic
       topic = this._parseString()
@@ -330,11 +356,29 @@ class Parser extends EventEmitter {
       if (this._pos >= packet.length) return this._emitError(new Error('Malformed Subscribe Payload'))
 
       options = this._parseByte()
+
+      if (this.settings.protocolVersion === 5) {
+        if (options & 0xc0) {
+          return this._emitError(new Error('Invalid subscribe topic flag bits, bits 7-6 must be 0'))
+        }
+      } else {
+        if (options & 0xfc) {
+          return this._emitError(new Error('Invalid subscribe topic flag bits, bits 7-2 must be 0'))
+        }
+      }
+
       qos = options & constants.SUBSCRIBE_OPTIONS_QOS_MASK
+      if (qos > 2) {
+        return this._emitError(new Error('Invalid subscribe QoS, must be <= 2'))
+      }
       nl = ((options >> constants.SUBSCRIBE_OPTIONS_NL_SHIFT) & constants.SUBSCRIBE_OPTIONS_NL_MASK) !== 0
       rap = ((options >> constants.SUBSCRIBE_OPTIONS_RAP_SHIFT) & constants.SUBSCRIBE_OPTIONS_RAP_MASK) !== 0
       rh = (options >> constants.SUBSCRIBE_OPTIONS_RH_SHIFT) & constants.SUBSCRIBE_OPTIONS_RH_MASK
 
+      if (rh > 2) {
+        return this._emitError(new Error('Invalid retain handling, must be <= 2'))
+      }
+
       subscription = { topic, qos }
 
       // mqtt 5 options
@@ -369,9 +413,21 @@ class Parser extends EventEmitter {
       }
     }
 
+    if (packet.length <= 0) { return this._emitError(new Error('Malformed suback, no payload specified')) }
+
     // Parse granted QoSes
     while (this._pos < this.packet.length) {
-      this.packet.granted.push(this._list.readUInt8(this._pos++))
+      const code = this._list.readUInt8(this._pos++)
+      if (this.settings.protocolVersion === 5) {
+        if (!constants.MQTT5_SUBACK_CODES[code]) {
+          return this._emitError(new Error('Invalid suback code'))
+        }
+      } else {
+        if (code > 2) {
+          return this._emitError(new Error('Invalid suback QoS, must be <= 2'))
+        }
+      }
+      this.packet.granted.push(code)
     }
   }
 
@@ -392,6 +448,8 @@ class Parser extends EventEmitter {
       }
     }
 
+    if (packet.length <= 0) { return this._emitError(new Error('Malformed unsubscribe, no payload specified')) }
+
     while (this._pos < packet.length) {
       // Parse topic
       const topic = this._parseString()
@@ -407,6 +465,13 @@ class Parser extends EventEmitter {
     debug('_parseUnsuback')
     const packet = this.packet
     if (!this._parseMessageId()) return this._emitError(new Error('Cannot parse messageId'))
+
+    if ((this.settings.protocolVersion === 3 ||
+      this.settings.protocolVersion === 4) && packet.length !== 2) {
+      return this._emitError(new Error('Malformed unsuback, payload length must be 2'))
+    }
+    if (packet.length <= 0) { return this._emitError(new Error('Malformed unsuback, no payload specified')) }
+
     // Properties mqtt 5
     if (this.settings.protocolVersion === 5) {
       const properties = this._parseProperties()
@@ -415,8 +480,13 @@ class Parser extends EventEmitter {
       }
       // Parse granted QoSes
       packet.granted = []
+
       while (this._pos < this.packet.length) {
-        this.packet.granted.push(this._list.readUInt8(this._pos++))
+        const code = this._list.readUInt8(this._pos++)
+        if (!constants.MQTT5_UNSUBACK_CODES[code]) {
+          return this._emitError(new Error('Invalid unsuback code'))
+        }
+        this.packet.granted.push(code)
       }
     }
   }
@@ -432,6 +502,20 @@ class Parser extends EventEmitter {
       if (packet.length > 2) {
         // response code
         packet.reasonCode = this._parseByte()
+        switch (this.packet.cmd) {
+          case 'puback':
+          case 'pubrec':
+            if (!constants.MQTT5_PUBACK_PUBREC_CODES[packet.reasonCode]) {
+              return this._emitError(new Error('Invalid ' + this.packet.cmd + ' reason code'))
+            }
+            break
+          case 'pubrel':
+          case 'pubcomp':
+            if (!constants.MQTT5_PUBREL_PUBCOMP_CODES[packet.reasonCode]) {
+              return this._emitError(new Error('Invalid ' + this.packet.cmd + ' reason code'))
+            }
+            break
+        }
         debug('_parseConfirmation: packet.reasonCode `%d`', packet.reasonCode)
       } else {
         packet.reasonCode = 0
@@ -458,6 +542,9 @@ class Parser extends EventEmitter {
       // response code
       if (this._list.length > 0) {
         packet.reasonCode = this._parseByte()
+        if (!constants.MQTT5_DISCONNECT_CODES[packet.reasonCode]) {
+          this._emitError(new Error('Invalid disconnect reason code'))
+        }
       } else {
         packet.reasonCode = 0
       }
@@ -483,6 +570,9 @@ class Parser extends EventEmitter {
 
     // response code
     packet.reasonCode = this._parseByte()
+    if (!constants.MQTT5_AUTH_CODES[packet.reasonCode]) {
+      return this._emitError(new Error('Invalid auth reason code'))
+    }
     // properies mqtt 5
     const properties = this._parseProperties()
     if (Object.getOwnPropertyNames(properties).length) {
@@ -590,20 +680,26 @@ class Parser extends EventEmitter {
       this._pos += bytes
     }
 
-    result = result
-      ? fullInfoFlag ? {
-        bytes,
-        value
-      } : value
-      : false
+    if (result) {
+      if (fullInfoFlag) {
+        result = { bytes, value }
+      } else {
+        result = value
+      }
+    } else {
+      result = false
+    }
 
     debug('_parseVarByteNum: result: %o', result)
     return result
   }
 
   _parseByte () {
-    const result = this._list.readUInt8(this._pos)
-    this._pos++
+    let result
+    if (this._pos < this._list.length) {
+      result = this._list.readUInt8(this._pos)
+      this._pos++
+    }
     debug('_parseByte: result: %o', result)
     return result
   }
@@ -646,6 +742,10 @@ class Parser extends EventEmitter {
     const result = {}
     while (this._pos < end) {
       const type = this._parseByte()
+      if (!type) {
+        this._emitError(new Error('Cannot parse property code type'))
+        return false
+      }
       const name = constants.propertiesCodes[type]
       if (!name) {
         this._emitError(new Error('Unknown property'))
