@@ -572,9 +572,9 @@ testParseGenerateDefaults('no clientId with 5.0', {
   clean: true,
   keepalive: 60,
   properties:
-    {
-      receiveMaximum: 20
-    },
+  {
+    receiveMaximum: 20
+  },
   clientId: ''
 }, Buffer.from(
   [16, 16, 0, 4, 77, 81, 84, 84, 5, 2, 0, 60, 3, 33, 0, 20, 0, 0]
@@ -1224,6 +1224,103 @@ testParseError('Cannot parse password', Buffer.from([
   0, 0 // Password
 ]))
 
+// Where a flag bit is marked as “Reserved” in Table 2.2 - Flag Bits, it is reserved for future use and MUST be set to the value listed in that table [MQTT-2.2.2-1]. If invalid flags are received, the receiver MUST close the Network Connection [MQTT-2.2.2-2]
+testParseError('Invalid header flag bits, must be 0x0 for connect packet', Buffer.from([
+  18, 10, // Header
+  0, 4, // Protocol ID length
+  0x4d, 0x51, 0x54, 0x54, // Protocol ID
+  3, // Protocol version
+  2, // Connect flags
+  0, 30 // Keepalive
+]))
+
+// The Server MUST validate that the reserved flag in the CONNECT Control Packet is set to zero and disconnect the Client if it is not zero [MQTT-3.1.2-3]
+testParseError('Connect flag bit 0 must be 0, but got 1', Buffer.from([
+  16, 10, // Header
+  0, 4, // Protocol ID length
+  0x4d, 0x51, 0x54, 0x54, // Protocol ID
+  3, // Protocol version
+  3, // Connect flags
+  0, 30 // Keepalive
+]))
+
+// If the Will Flag is set to 0 the Will QoS and Will Retain fields in the Connect Flags MUST be set to zero and the Will Topic and Will Message fields MUST NOT be present in the payload [MQTT-3.1.2-11].
+testParseError('Will Retain Flag must be set to zero when Will Flag is set to 0', Buffer.from([
+  16, 10, // Header
+  0, 4, // Protocol ID length
+  0x4d, 0x51, 0x54, 0x54, // Protocol ID
+  3, // Protocol version
+  0x22, // Connect flags
+  0, 30 // Keepalive
+]))
+
+// If the Will Flag is set to 0 the Will QoS and Will Retain fields in the Connect Flags MUST be set to zero and the Will Topic and Will Message fields MUST NOT be present in the payload [MQTT-3.1.2-11].
+testParseError('Will QoS must be set to zero when Will Flag is set to 0', Buffer.from([
+  16, 10, // Header
+  0, 4, // Protocol ID length
+  0x4d, 0x51, 0x54, 0x54, // Protocol ID
+  3, // Protocol version
+  0x12, // Connect flags
+  0, 30 // Keepalive
+]))
+
+// If the Will Flag is set to 0 the Will QoS and Will Retain fields in the Connect Flags MUST be set to zero and the Will Topic and Will Message fields MUST NOT be present in the payload [MQTT-3.1.2-11].
+testParseError('Will QoS must be set to zero when Will Flag is set to 0', Buffer.from([
+  16, 10, // Header
+  0, 4, // Protocol ID length
+  0x4d, 0x51, 0x54, 0x54, // Protocol ID
+  3, // Protocol version
+  0xa, // Connect flags
+  0, 30 // Keepalive
+]))
+
+// CONNECT, SUBSCRIBE, SUBACK, UNSUBSCRIBE, UNSUBACK (v.5) packets must have payload
+// CONNECT
+testParseError('Packet too short', Buffer.from([
+  16, // Header
+  8, // Packet length
+  0, 4, // Protocol ID length
+  77, 81, 84, 84, // MQTT
+  5, // Version
+  2, // Clean Start enabled
+  0, 0, // Keep-Alive
+  0, // Property Length
+  0, 0 // Properties
+  // No payload
+]), { protocolVersion: 5 })
+// SUBSCRIBE
+testParseError('Malformed subscribe, no payload specified', Buffer.from([
+  130, // Header
+  0 // Packet length
+]), { protocolVersion: 5 })
+// SUBACK
+testParseError('Malformed suback, no payload specified', Buffer.from([
+  144, // Header
+  0 // Packet length
+]), { protocolVersion: 5 })
+// UNSUBSCRIBE
+testParseError('Malformed unsubscribe, no payload specified', Buffer.from([
+  162, // Header
+  0 // Packet length
+]), { protocolVersion: 5 })
+// UNSUBACK (v.5)
+testParseError('Malformed unsuback, no payload specified', Buffer.from([
+  176, // Header
+  0 // Packet length
+]), { protocolVersion: 5 })
+// UNSUBACK (v.4)
+testParseError('Malformed unsuback, payload length must be 2', Buffer.from([
+  176, // Header
+  1, // Packet length
+  1
+]), { protocolVersion: 4 })
+// UNSUBACK (v.3)
+testParseError('Malformed unsuback, payload length must be 2', Buffer.from([
+  176, // Header
+  1, // Packet length
+  1
+]), { protocolVersion: 3 })
+
 testParseGenerate('connack with return code 0', {
   cmd: 'connack',
   retain: false,
@@ -1363,6 +1460,20 @@ testParseGenerate('connack with return code 5', {
   32, 2, 0, 5
 ]))
 
+// Where a flag bit is marked as “Reserved” in Table 2.2 - Flag Bits, it is reserved for future use and MUST be set to the value listed in that table [MQTT-2.2.2-1]. If invalid flags are received, the receiver MUST close the Network Connection [MQTT-2.2.2-2]
+testParseError('Invalid header flag bits, must be 0x0 for connack packet', Buffer.from([
+  33, 2, // header
+  0, // flags
+  5 // return code
+]))
+
+// Byte 1 is the "Connect Acknowledge Flags". Bits 7-1 are reserved and MUST be set to 0 [MQTT-3.2.2-1].
+testParseError('Invalid connack flags, bits 7-1 must be set to 0', Buffer.from([
+  32, 2, // header
+  2, // flags
+  5 // return code
+]))
+
 testGenerateError('Invalid return code', {
   cmd: 'connack',
   retain: false,
@@ -1609,6 +1720,14 @@ testParseGenerate('empty publish', {
   // Empty payload
 ]))
 
+// A PUBLISH Packet MUST NOT have both QoS bits set to 1. If a Server or Client receives a PUBLISH Packet which has both QoS bits set to 1 it MUST close the Network Connection [MQTT-3.3.1-4].
+testParseError('Packet must not have both QoS bits set to 1', Buffer.from([
+  0x36, 6, // Header
+  0, 4, // Topic length
+  116, 101, 115, 116 // Topic
+  // Empty payload
+]))
+
 test('splitted publish parse', t => {
   t.plan(3)
 
@@ -1744,6 +1863,12 @@ testParseGenerate('puback', {
   0, 2 // Message ID
 ]))
 
+// Where a flag bit is marked as “Reserved” in Table 2.2 - Flag Bits, it is reserved for future use and MUST be set to the value listed in that table [MQTT-2.2.2-1]. If invalid flags are received, the receiver MUST close the Network Connection [MQTT-2.2.2-2]
+testParseError('Invalid header flag bits, must be 0x0 for puback packet', Buffer.from([
+  65, 2, // Header
+  0, 2 // Message ID
+]))
+
 testParseGenerate('puback with reason and no MQTT 5 properties', {
   cmd: 'puback',
   retain: false,
@@ -1781,6 +1906,13 @@ testParseGenerate('puback MQTT 5 properties', {
   38, 0, 4, 116, 101, 115, 116, 0, 4, 116, 101, 115, 116 // userProperties
 ]), { protocolVersion: 5 })
 
+testParseError('Invalid puback reason code', Buffer.from([
+  64, 4, // Header
+  0, 2, // Message ID
+  0x11, // reason code
+  0 // properties length
+]), { protocolVersion: 5 })
+
 testParseGenerate('pubrec', {
   cmd: 'pubrec',
   retain: false,
@@ -1793,6 +1925,12 @@ testParseGenerate('pubrec', {
   0, 2 // Message ID
 ]))
 
+// Where a flag bit is marked as “Reserved” in Table 2.2 - Flag Bits, it is reserved for future use and MUST be set to the value listed in that table [MQTT-2.2.2-1]. If invalid flags are received, the receiver MUST close the Network Connection [MQTT-2.2.2-2]
+testParseError('Invalid header flag bits, must be 0x0 for pubrec packet', Buffer.from([
+  81, 2, // Header
+  0, 2 // Message ID
+]))
+
 testParseGenerate('pubrec MQTT 5 properties', {
   cmd: 'pubrec',
   retain: false,
@@ -1828,6 +1966,19 @@ testParseGenerate('pubrel', {
   0, 2 // Message ID
 ]))
 
+testParseError('Invalid pubrel reason code', Buffer.from([
+  98, 4, // Header
+  0, 2, // Message ID
+  0x11, // Reason code
+  0 // Properties length
+]), { protocolVersion: 5 })
+
+// Where a flag bit is marked as “Reserved” in Table 2.2 - Flag Bits, it is reserved for future use and MUST be set to the value listed in that table [MQTT-2.2.2-1]. If invalid flags are received, the receiver MUST close the Network Connection [MQTT-2.2.2-2]
+testParseError('Invalid header flag bits, must be 0x2 for pubrel packet', Buffer.from([
+  96, 2, // Header
+  0, 2 // Message ID
+]))
+
 testParseGenerate('pubrel MQTT5 properties', {
   cmd: 'pubrel',
   retain: false,
@@ -1835,7 +1986,7 @@ testParseGenerate('pubrel MQTT5 properties', {
   dup: false,
   length: 24,
   messageId: 2,
-  reasonCode: 16,
+  reasonCode: 0x92,
   properties: {
     reasonString: 'test',
     userProperties: {
@@ -1845,12 +1996,19 @@ testParseGenerate('pubrel MQTT5 properties', {
 }, Buffer.from([
   98, 24, // Header
   0, 2, // Message ID
-  16, // reason code
+  0x92, // reason code
   20, // properties length
   31, 0, 4, 116, 101, 115, 116, // reasonString
   38, 0, 4, 116, 101, 115, 116, 0, 4, 116, 101, 115, 116 // userProperties
 ]), { protocolVersion: 5 })
 
+testParseError('Invalid pubrel reason code', Buffer.from([
+  98, 4, // Header
+  0, 2, // Message ID
+  16, // reason code
+  0 // properties length
+]), { protocolVersion: 5 })
+
 testParseGenerate('pubcomp', {
   cmd: 'pubcomp',
   retain: false,
@@ -1863,6 +2021,12 @@ testParseGenerate('pubcomp', {
   0, 2 // Message ID
 ]))
 
+// Where a flag bit is marked as “Reserved” in Table 2.2 - Flag Bits, it is reserved for future use and MUST be set to the value listed in that table [MQTT-2.2.2-1]. If invalid flags are received, the receiver MUST close the Network Connection [MQTT-2.2.2-2]
+testParseError('Invalid header flag bits, must be 0x0 for pubcomp packet', Buffer.from([
+  113, 2, // Header
+  0, 2 // Message ID
+]))
+
 testParseGenerate('pubcomp MQTT 5 properties', {
   cmd: 'pubcomp',
   retain: false,
@@ -1870,7 +2034,7 @@ testParseGenerate('pubcomp MQTT 5 properties', {
   dup: false,
   length: 24,
   messageId: 2,
-  reasonCode: 16,
+  reasonCode: 0x92,
   properties: {
     reasonString: 'test',
     userProperties: {
@@ -1880,13 +2044,20 @@ testParseGenerate('pubcomp MQTT 5 properties', {
 }, Buffer.from([
   112, 24, // Header
   0, 2, // Message ID
-  16, // reason code
+  0x92, // reason code
   20, // properties length
   31, 0, 4, 116, 101, 115, 116, // reasonString
   38, 0, 4, 116, 101, 115, 116, 0, 4, 116, 101, 115, 116 // userProperties
 ]), { protocolVersion: 5 })
 
-testParseError('Wrong subscribe header', Buffer.from([
+testParseError('Invalid pubcomp reason code', Buffer.from([
+  112, 4, // Header
+  0, 2, // Message ID
+  16, // reason code
+  0 // properties length
+]), { protocolVersion: 5 })
+
+testParseError('Invalid header flag bits, must be 0x2 for subscribe packet', Buffer.from([
   128, 9, // Header (subscribeqos=0length=9)
   0, 6, // Message ID (6)
   0, 4, // Topic length,
@@ -1915,6 +2086,40 @@ testParseGenerate('subscribe to one topic', {
   0 // Qos (0)
 ]))
 
+testParseError('Invalid subscribe QoS, must be <= 2', Buffer.from([
+  130, 9, // Header (subscribeqos=0length=9)
+  0, 6, // Message ID (6)
+  0, 4, // Topic length,
+  116, 101, 115, 116, // Topic (test)
+  3 // Qos
+]))
+
+testParseError('Invalid subscribe topic flag bits, bits 7-6 must be 0', Buffer.from([
+  130, 10, // Header (subscribeqos=0length=9)
+  0, 6, // Message ID (6)
+  0, // Property length (0)
+  0, 4, // Topic length,
+  116, 101, 115, 116, // Topic (test)
+  0x80 // Flags
+]), { protocolVersion: 5 })
+
+testParseError('Invalid retain handling, must be <= 2', Buffer.from([
+  130, 10, // Header (subscribeqos=0length=9)
+  0, 6, // Message ID (6)
+  0, // Property length (0)
+  0, 4, // Topic length,
+  116, 101, 115, 116, // Topic (test)
+  0x30 // Flags
+]), { protocolVersion: 5 })
+
+testParseError('Invalid subscribe topic flag bits, bits 7-2 must be 0', Buffer.from([
+  130, 9, // Header (subscribeqos=0length=9)
+  0, 6, // Message ID (6)
+  0, 4, // Topic length,
+  116, 101, 115, 116, // Topic (test)
+  0x08 // Flags
+]))
+
 testParseGenerate('subscribe to one topic by MQTT 5', {
   cmd: 'subscribe',
   retain: false,
@@ -2038,15 +2243,42 @@ testParseGenerate('suback', {
   retain: false,
   qos: 0,
   dup: false,
-  length: 6,
+  length: 5,
+  granted: [0, 1, 2],
+  messageId: 6
+}, Buffer.from([
+  144, 5, // Header
+  0, 6, // Message ID
+  0, 1, 2
+]))
+
+testParseGenerate('suback', {
+  cmd: 'suback',
+  retain: false,
+  qos: 0,
+  dup: false,
+  length: 7,
   granted: [0, 1, 2, 128],
   messageId: 6
 }, Buffer.from([
+  144, 7, // Header
+  0, 6, // Message ID
+  0, // Property length
+  0, 1, 2, 128 // Granted qos (0, 1, 2) and a rejected being 0x80
+]), { protocolVersion: 5 })
+
+testParseError('Invalid suback QoS, must be <= 2', Buffer.from([
   144, 6, // Header
   0, 6, // Message ID
   0, 1, 2, 128 // Granted qos (0, 1, 2) and a rejected being 0x80
 ]))
 
+testParseError('Invalid suback code', Buffer.from([
+  144, 6, // Header
+  0, 6, // Message ID
+  0, 1, 2, 0x79 // Granted qos (0, 1, 2) and an invalid code
+]), { protocolVersion: 5 })
+
 testParseGenerate('suback MQTT 5', {
   cmd: 'suback',
   retain: false,
@@ -2090,6 +2322,16 @@ testParseGenerate('unsubscribe', {
   116, 101, 115, 116 // Topic (test)
 ]))
 
+// Where a flag bit is marked as “Reserved” in Table 2.2 - Flag Bits, it is reserved for future use and MUST be set to the value listed in that table [MQTT-2.2.2-1]. If invalid flags are received, the receiver MUST close the Network Connection [MQTT-2.2.2-2]
+testParseError('Invalid header flag bits, must be 0x2 for unsubscribe packet', Buffer.from([
+  160, 14,
+  0, 7, // Message ID (7)
+  0, 4, // Topic length
+  116, 102, 115, 116, // Topic (tfst)
+  0, 4, // Topic length,
+  116, 101, 115, 116 // Topic (test)
+]))
+
 testGenerateError('Invalid unsubscriptions', {
   cmd: 'unsubscribe',
   retain: false,
@@ -2149,6 +2391,12 @@ testParseGenerate('unsuback', {
   0, 8 // Message ID
 ]))
 
+// Where a flag bit is marked as “Reserved” in Table 2.2 - Flag Bits, it is reserved for future use and MUST be set to the value listed in that table [MQTT-2.2.2-1]. If invalid flags are received, the receiver MUST close the Network Connection [MQTT-2.2.2-2]
+testParseError('Invalid header flag bits, must be 0x0 for unsuback packet', Buffer.from([
+  177, 2, // Header
+  0, 8 // Message ID
+]))
+
 testParseGenerate('unsuback MQTT 5', {
   cmd: 'unsuback',
   retain: false,
@@ -2172,6 +2420,13 @@ testParseGenerate('unsuback MQTT 5', {
   0, 128 // success and error
 ]), { protocolVersion: 5 })
 
+testParseError('Invalid unsuback code', Buffer.from([
+  176, 4, // Header
+  0, 8, // Message ID
+  0, // properties length
+  0x84 // reason codes
+]), { protocolVersion: 5 })
+
 testParseGenerate('pingreq', {
   cmd: 'pingreq',
   retain: false,
@@ -2182,6 +2437,11 @@ testParseGenerate('pingreq', {
   192, 0 // Header
 ]))
 
+// Where a flag bit is marked as “Reserved” in Table 2.2 - Flag Bits, it is reserved for future use and MUST be set to the value listed in that table [MQTT-2.2.2-1]. If invalid flags are received, the receiver MUST close the Network Connection [MQTT-2.2.2-2]
+testParseError('Invalid header flag bits, must be 0x0 for pingreq packet', Buffer.from([
+  193, 0 // Header
+]))
+
 testParseGenerate('pingresp', {
   cmd: 'pingresp',
   retain: false,
@@ -2192,6 +2452,11 @@ testParseGenerate('pingresp', {
   208, 0 // Header
 ]))
 
+// Where a flag bit is marked as “Reserved” in Table 2.2 - Flag Bits, it is reserved for future use and MUST be set to the value listed in that table [MQTT-2.2.2-1]. If invalid flags are received, the receiver MUST close the Network Connection [MQTT-2.2.2-2]
+testParseError('Invalid header flag bits, must be 0x0 for pingresp packet', Buffer.from([
+  209, 0 // Header
+]))
+
 testParseGenerate('disconnect', {
   cmd: 'disconnect',
   retain: false,
@@ -2202,6 +2467,11 @@ testParseGenerate('disconnect', {
   224, 0 // Header
 ]))
 
+// Where a flag bit is marked as “Reserved” in Table 2.2 - Flag Bits, it is reserved for future use and MUST be set to the value listed in that table [MQTT-2.2.2-1]. If invalid flags are received, the receiver MUST close the Network Connection [MQTT-2.2.2-2]
+testParseError('Invalid header flag bits, must be 0x0 for disconnect packet', Buffer.from([
+  225, 0 // Header
+]))
+
 testParseGenerate('disconnect MQTT 5', {
   cmd: 'disconnect',
   retain: false,
@@ -2240,6 +2510,12 @@ testParseGenerate('disconnect MQTT 5 with no properties', {
   0 // Property Length (0 => No Properties)
 ]), { protocolVersion: 5 })
 
+testParseError('Invalid disconnect reason code', Buffer.from([
+  224, 2, // Fixed Header (DISCONNECT, Remaining Length)
+  0x05, // Reason Code (Normal Disconnection)
+  0 // Property Length (0 => No Properties)
+]), { protocolVersion: 5 })
+
 testParseGenerate('auth MQTT 5', {
   cmd: 'auth',
   retain: false,
@@ -2265,6 +2541,12 @@ testParseGenerate('auth MQTT 5', {
   38, 0, 4, 116, 101, 115, 116, 0, 4, 116, 101, 115, 116 // userProperties
 ]), { protocolVersion: 5 })
 
+testParseError('Invalid auth reason code', Buffer.from([
+  240, 2, // Fixed Header (DISCONNECT, Remaining Length)
+  0x17, // Reason Code
+  0 // Property Length (0 => No Properties)
+]), { protocolVersion: 5 })
+
 testGenerateError('Invalid protocolId', {
   cmd: 'connect',
   retain: false,
@@ -2690,6 +2972,27 @@ testParseError('Malformed Subscribe Payload', Buffer.from([
   0 // requested QoS
 ]))
 
+test('Cannot parse property code type', t => {
+  const packets = Buffer.from([
+    16, 16, 0, 4, 77, 81, 84, 84, 5, 2, 0, 60, 3, 33, 0, 20, 0, 0, 98, 2, 211, 1, 224, 2, 0, 32
+  ])
+
+  t.plan(3)
+
+  const parser = mqtt.parser()
+
+  parser.on('error', err => {
+    t.equal(err.message, 'Cannot parse property code type', 'expected error message')
+    t.end()
+  })
+
+  parser.on('packet', (packet) => {
+    t.pass('Packet parsed')
+  })
+
+  parser.parse(packets)
+})
+
 testWriteToStreamError('Invalid command', {
   cmd: 'invalid'
 })
