@@ -83,7 +83,7 @@ function uncork (stream) {
 function connect (packet, stream, opts) {
   const settings = packet || {}
   const protocolId = settings.protocolId || 'MQTT'
-  const protocolVersion = settings.protocolVersion || 4
+  let protocolVersion = settings.protocolVersion || 4
   const will = settings.will
   let clean = settings.clean
   const keepalive = settings.keepalive || 0
@@ -110,10 +110,10 @@ function connect (packet, stream, opts) {
     return false
   } else length += 1
 
-  // ClientId might be omitted in 3.1.1, but only if cleanSession is set to 1
+  // ClientId might be omitted in 3.1.1 and 5, but only if cleanSession is set to 1
   if ((typeof clientId === 'string' || Buffer.isBuffer(clientId)) &&
-     (clientId || protocolVersion === 4) && (clientId || clean)) {
-    length += clientId.length + 2
+     (clientId || protocolVersion >= 4) && (clientId || clean)) {
+    length += Buffer.byteLength(clientId) + 2
   } else {
     if (protocolVersion < 4) {
       stream.emit('error', new Error('clientId must be supplied before 3.1.1'))
@@ -137,9 +137,12 @@ function connect (packet, stream, opts) {
   // Connect flags
   length += 1
 
+  let propertiesData
+  let willProperties
+
   // Properties
   if (protocolVersion === 5) {
-    var propertiesData = getProperties(stream, properties)
+    propertiesData = getProperties(stream, properties)
     if (!propertiesData) { return false }
     length += propertiesData.length
   }
@@ -174,7 +177,7 @@ function connect (packet, stream, opts) {
       }
     }
     // will properties
-    var willProperties = {}
+    willProperties = {}
     if (protocolVersion === 5) {
       willProperties = getProperties(stream, will.properties)
       if (!willProperties) { return false }
@@ -217,12 +220,21 @@ function connect (packet, stream, opts) {
 
   // Generate protocol ID
   writeStringOrBuffer(stream, protocolId)
+
+  if (settings.bridgeMode) {
+    protocolVersion += 128
+  }
+
   stream.write(
-    protocolVersion === 4
-      ? protocol.VERSION4
-      : protocolVersion === 5
-        ? protocol.VERSION5
-        : protocol.VERSION3
+    protocolVersion === 131
+      ? protocol.VERSION131
+      : protocolVersion === 132
+        ? protocol.VERSION132
+        : protocolVersion === 4
+          ? protocol.VERSION4
+          : protocolVersion === 5
+            ? protocol.VERSION5
+            : protocol.VERSION3
   )
 
   // Connect flags
@@ -780,6 +792,11 @@ function auth (packet, stream, opts) {
 
 const varByteIntCache = {}
 function writeVarByteInt (stream, num) {
+  if (num > protocol.VARBYTEINT_MAX) {
+    stream.emit('error', new Error(`Invalid variable byte integer: ${num}`))
+    return false
+  }
+
   let buffer = varByteIntCache[num]
 
   if (!buffer) {
@@ -787,7 +804,7 @@ function writeVarByteInt (stream, num) {
     if (num < 16384) varByteIntCache[num] = buffer
   }
   debug('writeVarByteInt: writing to stream: %o', buffer)
-  stream.write(buffer)
+  return stream.write(buffer)
 }
 
 /**
